import matplotlib.pyplot as plt

from ngsolve import *
import netgen.meshing as ng
import numpy as np


class AnalyticSolution(object):

    def __init__(self, *args, **kwargs):
        """
        Calculate the analytical solution for one side heated plate.

        See: The 1-D Heat Equation Matthew J. Hancock S 9

        :param args:
        :key thickness: thickness of the plate in m
        :key density: density of the plate in in kg/m³
        :key cp: specific heat capacity in J/kg K
        :key heat_conductivity: heat conductivity in W/m K
        :key t_inf: temperature of environment
        :key t_0: initial temperature
        """

        self.thickness = kwargs.get('thickness')
        self.density = kwargs.get('density')  # density in kg/m³
        self.cp = kwargs.get('cp')  # specific heat capacity in J/kg K
        self.heat_conductivity = kwargs.get('heat_conductivity')  # heat conductivity in W/m K

        self.t_inf = kwargs.get('t_inf')  # temperature of environment
        self.t_0 = kwargs.get('t_0')  # initial temperature

        self.thermal_diffusivity = self.heat_conductivity / (self.density * self.cp)  # mm²/s

        self.n = 1000

    def x(self, x):
        return x/self.thickness

    def t(self, t, x):

        t = self.thermal_diffusivity * t / (self.thickness ** 2)

        n_sum = 0
        x = self.x(x)

        for n in range(1, self.n):
            n_sum += np.sin((2 * n - 1) * np.pi * x) / (2 * n - 1) * np.exp(-(2 * n - 1) ** 2 * np.pi ** 2 * t)

        return self.t_inf + 4 * (self.t_0 - self.t_inf) / np.pi * n_sum


# Define range of densities in the regions   [kg/m^3]
region_densities = np.array([1360])  # Frame

# Define range of specific heat capacities in the regions  [J/K kg]
region_hcs = np.array([1170])

# Define range of thermal conductivities in the regions [J/msK]
region_conductivities = np.array([100])

region_diffusivity = region_conductivities / (region_densities * region_hcs) * 1e6


# Initial value of solution in °C
initial_temperature = 20

# timestep size in s
dt = 1

# final time in s
t_end = 1_000


"""1. generate a 1D mesh"""

mesh = ng.Mesh()
mesh.dim = 1
nel = 20
pnums = []

# add nodes and ID
for i in range(nel+1):
    node = ng.Pnt(i/nel, 0, 0)
    mesh.Add(ng.MeshPoint(node))
    pnums.append(i+1)  # node index starts 1

for i in range(nel):
    mesh.Add(ng.Element1D([pnums[i], pnums[i+1]], index=1))

# add BC points
mesh.Add(ng.Element0D(pnums[0], index=1))
mesh.Add(ng.Element0D(pnums[nel], index=2))

# set boundary condition names
mesh.SetBCName(0, 'left')
mesh.SetBCName(1, 'right')

print(pnums)
mesh = Mesh(mesh)
print(mesh.GetBoundaries())

"""2. compute on this mesh"""

print(mesh.GetBoundaries())

K = CoefficientFunction(region_conductivities)

fes = H1(mesh, order=1, dirichlet='left|right')
gfuDir = GridFunction(fes)

u = fes.TrialFunction()  # symbolic object
v = fes.TestFunction()  # symbolic object
gfu = GridFunction(fes)  # solution

a = BilinearForm(fes, symmetric=False)
a += SymbolicBFI(K * grad(u) * grad(v))
a.Assemble()

m = BilinearForm(fes, symmetric=False)
m += SymbolicBFI(CoefficientFunction(region_densities * region_conductivities) * u * v)
m.Assemble()

mstar = m.mat.CreateMatrix()
mstar.AsVector().data = m.mat.AsVector() + dt * a.mat.AsVector()
invmstar = mstar.Inverse(freedofs=fes.FreeDofs(), inverse="sparsecholesky")
res = gfu.vec.CreateVector()

f = LinearForm(fes)
f.Assemble()

# set dirichlet boundary conditions:
gfu.vec.FV().NumPy()[[0, nel]] = initial_temperature


# #######################################
# Start solving time steps:
# #######################################

time = 0

times = np.array([time])
solutions = gfu.vec.FV().NumPy()

while (time + dt) <= t_end:
    # print(f'calculating time {self.time}')
    time = round(time + dt, 6)
    res.data = f.vec - a.mat * gfu.vec
    gfu.vec.data += dt * invmstar * res

    # add to solutions / time array
    solutions = np.vstack((solutions, gfu.vec.data))
    times = np.vstack((times, time))

print('solution calculated')

# ##############################################################################
# plot results for comparison
# ##############################################################################

x = np.linspace(0, 1, 100)
t_eval = [10, 100, 500, 1_000]
node_positions = [x.point for x in mesh.vertices]

# analytic solution:

analytic_solution = AnalyticSolution(thickness=1,
                                     density=region_densities[0],
                                     cp=region_hcs[0],
                                     heat_conductivity=region_conductivities[0],
                                     t_inf=initial_temperature,
                                     t_0=0)

f, axes = plt.subplots(1, t_eval.__len__(), sharey=True)


for i, t in enumerate(t_eval):
    indx = np.where(times[:, 0] == t)[0]
    axes[i].plot(np.array(node_positions)[:, 0], solutions[indx, :][0, :], label='numerical')
    axes[i].plot(x, analytic_solution.t(t, x), label='analytic')
    axes[i].legend()
    axes[i].set_title(f't = {t} s')
plt.show()
